""" This module contains functions to read spectra and save deltas. Can be
imported without a need for MPI."""

import argparse
import warnings

import fitsio
import numpy as np

from qsonic import QsonicException
import qsonic.spectrum


def add_io_parser(parser=None):
    """ Adds IO related arguments to parser. These arguments are grouped under
    'Input/output parameters and selections'. ``--input-dir`` and ``--catalog``
    are required.

    Arguments
    ---------
    parser: argparse.ArgumentParser, default: None

    Returns
    ---------
    parser: argparse.ArgumentParser
    """
    if parser is None:
        parser = argparse.ArgumentParser(
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    ingroup = parser.add_argument_group('Input options')
    ingroup.add_argument(
        "--input-dir", '-i', required=True,
        help="Input directory.")
    ingroup.add_argument(
        "--catalog", required=True,
        help="Catalog filename")
    ingroup.add_argument(
        "--mock-analysis", action="store_true",
        help="Input folder is mock. Uses nside=16")
    ingroup.add_argument(
        "--keep-surveys", nargs='+', default=['sv3', 'main'],
        help="Surveys to keep.")
    ingroup.add_argument(
        "--skip-resomat", action="store_true",
        help="Skip reading resolution matrix for 3D.")
    ingroup.add_argument(
        "--arms", default=['B', 'R'], choices=['B', 'R', 'Z'], nargs='+',
        help="Arms to read.")

    outgroup = parser.add_argument_group('Output options')
    outgroup.add_argument(
        "--outdir", '-o',
        help="Output directory to save deltas.")
    outgroup.add_argument(
        "--coadd-arms", action="store_true",
        help="Coadds arms when saving.")
    outgroup.add_argument(
        "--save-smooth-weights", action="store_true",
        help="Save smoothed weights instead.")
    outgroup.add_argument(
        "--save-by-hpx", action="store_true",
        help="Save by healpix. If not, saves by MPI rank.")
    return parser


def read_spectra_onehealpix(
        catalog_hpx, input_dir, arms_to_keep, mock_analysis, skip_resomat,
        program="dark"):
    """ Returns a list of Spectrum objects for a given catalog of a single
    healpix.

    Arguments
    ---------
    catalog_hpx: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Catalog of quasars in a single healpix. If for data 'SURVEY' column
        must be present and sorted.
    input_dir: str
        Input directory
    arms_to_keep: list(str)
        Must only contain B, R and Z
    mock_analysis: bool
        Reads for mock data if true.
    skip_resomat: bool
        If true, do not read resomat.
    program: str
        Always use dark program.

    Returns
    ---------
    spectra_list: list(Spectrum)

    Raises
    ---------
    RuntimeWarning
        If number of quasars in the healpix file does not match the catalog.
    """
    spectra_list = []

    if mock_analysis:
        data, idx_cat = read_onehealpix_file_mock(
            catalog_hpx, input_dir, arms_to_keep, skip_resomat
        )

        if idx_cat.size != catalog_hpx.size:
            catalog_hpx = catalog_hpx[idx_cat]

        spectra_list.extend(
            qsonic.spectrum.generate_spectra_list_from_data(
                catalog_hpx, data)
        )
    else:
        # Assume sorted by survey
        # cat.sort(order='SURVEY')
        unique_surveys, s2 = np.unique(
            catalog_hpx['SURVEY'], return_index=True)
        survey_split_cat = np.split(catalog_hpx, s2[1:])

        for cat_by_survey in survey_split_cat:
            data, idx_cat = read_onehealpix_file_data(
                cat_by_survey, input_dir, arms_to_keep,
                skip_resomat, program
            )

            if idx_cat.size != cat_by_survey.size:
                cat_by_survey = cat_by_survey[idx_cat]

            spectra_list.extend(
                qsonic.spectrum.generate_spectra_list_from_data(
                    cat_by_survey, data
                )
            )

    return spectra_list


def read_deltas(fname):
    """ Returns a list of all Delta objects in a file.

    Arguments
    ---------
    fname: str
        FITS file name

    Returns
    ---------
    deltas_list: list(Delta)

    Raises
    ---------
    RuntimeError
        If the file is missing certain columns and keys. See :class:`Delta`.
    """
    deltas_list = []
    fitsfile = fitsio.FITS(fname)

    for hdu in fitsfile[1:]:
        deltas_list.append(qsonic.spectrum.Delta(hdu))

    fitsfile.close()

    return deltas_list


def save_deltas(
        spectra_list, outdir, varlss_interp,
        save_by_hpx=False, mpi_rank=None, use_ivar_sm=False
):
    """ Saves given list of spectra as deltas. NO coaddition of arms.
    Each arm is saved separately. Only valid spectra are saved.

    Arguments
    ---------
    spectra_list: list(Spectrum)
        Continuum fitted spectra objects. All must be valid!
    outdir: str
        Output directory. Does not save if empty of None
    varlss_interp: Interpolator
        Interpolator for LSS variance
    save_by_hpx: bool, default: False
        Saves by healpix if True. Has priority over mpi_rank
    mpi_rank: int, default: None
        Rank of the MPI process. Save by `mpi_rank` if passed.
    use_ivar_sm: bool, default: False
        Use :attr:`Spectrum.forestivar_sm` in weights instead.

    Raises
    ---------
    QsonicException
        If both `mpi_rank` and `save_by_hpx` is None/False.
    QsonicException
        If blinding is not set.
    """
    if not outdir:
        return

    if save_by_hpx:
        pixnos = np.array([spec.hpix for spec in spectra_list])
        sort_idx = np.argsort(pixnos)
        pixnos = pixnos[sort_idx]
        unique_pix, s = np.unique(pixnos, return_index=True)
        split_spectra = np.split(np.array(spectra_list)[sort_idx], s[1:])
    elif mpi_rank is not None:
        unique_pix = [mpi_rank]
        split_spectra = [spectra_list]
    else:
        raise QsonicException("save_by_hpx and mpi_rank can't both be None.")

    if qsonic.spectrum.Spectrum.blinding_not_set():
        raise QsonicException("Blinding is not set. Cannot save delta.")

    for healpix, hp_specs in zip(unique_pix, split_spectra):
        results = fitsio.FITS(
            f"{outdir}/delta-{healpix}.fits", 'rw', clobber=True)

        for spec in qsonic.spectrum.valid_spectra(hp_specs):
            spec.write(results, varlss_interp, use_ivar_sm)

        results.close()


def _read_resoimage(imhdu, quasar_indices, nwave):
    # Reading into sorted order is faster
    # Since the output catalog is already sorted, we place them in order
    ndiags = imhdu.get_dims()[1]
    dtype = imhdu._get_image_numpy_dtype()
    data = np.empty((quasar_indices.size, ndiags, nwave), dtype=dtype)
    for idata, iqso in enumerate(quasar_indices):
        data[idata, :, :] = imhdu[int(iqso), :, :]

    return data


def _read_imagehdu(imhdu, quasar_indices, nwave):
    dtype = imhdu._get_image_numpy_dtype()
    data = np.empty((quasar_indices.size, nwave), dtype=dtype)
    for idata, iqso in enumerate(quasar_indices):
        data[idata, :] = imhdu[int(iqso), :]

    return data

# Timing showed this is slower
# def _read_imagehdu_2(imhdu, quasar_indices):
#     ndims = imhdu.get_info()['ndims']
#     if ndims == 2:
#         return np.vstack([imhdu[int(idx), :] for idx in quasar_indices])
#     return np.vstack([imhdu[int(idx), :, :] for idx in quasar_indices])


def _read_onehealpix_file(
        targetids_by_survey, fspec, arms_to_keep, skip_resomat):
    """ Common function to read a single fits file.

    Arguments
    ---------
    targetids_by_survey: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Targetids_by_survey (used to be catalog). If data, split by survey and
        contains only one survey.
    fspec: str
        Filename to open.
    arms_to_keep: list(str)
        Must only contain B, R and Z.
    skip_resomat: bool
        If true, do not read resomat.

    Returns
    ---------
    data: dict( :external+numpy:py:class:`ndarray <numpy.ndarray>` )
        Only quasar spectra are read into keywords wave, flux etc. Resolution
        is read if present.
    idx_cat: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Indices in `targetids_by_survey` there were succesfully read.

    Raises
    ---------
    RuntimeWarning
        If number of quasars in the healpix file does not match the catalog.
    """
    # Assume it is sorted
    # cat_by_survey.sort(order='TARGETID')
    fitsfile = fitsio.FITS(fspec)

    fbrmap = fitsfile['FIBERMAP'].read(columns='TARGETID')
    common_targetids, idx_fbr, idx_cat = np.intersect1d(
        fbrmap, targetids_by_survey, assume_unique=True, return_indices=True)
    if (common_targetids.size != targetids_by_survey.size):
        warnings.warn(
            f"Error reading {fspec}. "
            "Number of quasars in healpix does not match the catalog "
            f"catalog:{targetids_by_survey.size} vs "
            f"healpix:{common_targetids.size}!", RuntimeWarning)

    fbrmap = fbrmap[idx_fbr]

    data = {
        'wave': {},
        'flux': {},
        'ivar': {},
        'mask': {},
        'reso': {}
    }

    for arm in arms_to_keep:
        # Cannot read by rows= argument.
        data['wave'][arm] = fitsfile[f'{arm}_WAVELENGTH'].read()
        nwave = data['wave'][arm].size

        data['flux'][arm] = _read_imagehdu(
            fitsfile[f'{arm}_FLUX'], idx_fbr, nwave)
        data['ivar'][arm] = _read_imagehdu(
            fitsfile[f'{arm}_IVAR'], idx_fbr, nwave)
        data['mask'][arm] = _read_imagehdu(
            fitsfile[f'{arm}_MASK'], idx_fbr, nwave)

        if skip_resomat or f'{arm}_RESOLUTION' not in fitsfile:
            continue

        data['reso'][arm] = _read_resoimage(
            fitsfile[f'{arm}_RESOLUTION'], idx_fbr, nwave)

    fitsfile.close()

    return data, idx_cat


def read_onehealpix_file_data(
        cat_by_survey, input_dir, arms_to_keep, skip_resomat, program="dark"):
    """ Read a single fits file for data.

    Arguments
    ---------
    cat_by_survey: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Catalog for a single survey and healpix. Ordered by TARGETID.
    input_dir: str
        Input directory.
    arms_to_keep: list(str)
        Must only contain B, R and Z.
    skip_resomat: bool
        If true, do not read resomat.

    Returns
    ---------
    data: dict
        Only quasar spectra are read into keywords wave, flux etc. Resolution
        is read if present.
    idx_cat: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Indices in `cat_by_survey` there were succesfully read.

    Raises
    ---------
    RuntimeWarning
        If number of quasars in the healpix file does not match the catalog.
    """
    survey = cat_by_survey['SURVEY'][0]
    pixnum = cat_by_survey['HPXPIXEL'][0]

    fspec = (f"{input_dir}/{survey}/{program}/{pixnum//100}/"
             f"{pixnum}/coadd-{survey}-{program}-{pixnum}.fits")
    data, idx_cat = _read_onehealpix_file(
        cat_by_survey['TARGETID'], fspec, arms_to_keep, skip_resomat)

    return data, idx_cat


def read_onehealpix_file_mock(
        catalog_hpx, input_dir, arms_to_keep, skip_resomat, nside=16):
    """ Read a single FITS file for mocks.

    Arguments
    ---------
    catalog_hpx: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Catalog for a single healpix. Ordered by TARGETID.
    input_dir: str
        Input directory.
    arms_to_keep: list(str)
        Must only contain B, R and Z.
    skip_resomat: bool
        If true, do not read resomat.

    Returns
    ---------
    data: dict
        Only quasar spectra are read into keywords wave, flux etc. Resolution
        is read if present.
    idx_cat: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Indices in `catalog_hpx` there were succesfully read.

    Raises
    ---------
    RuntimeWarning
        If number of quasars in the healpix file does not match the catalog.
    """
    pixnum = catalog_hpx['HPXPIXEL'][0]
    fspec = f"{input_dir}/{pixnum//100}/{pixnum}/spectra-{nside}-{pixnum}.fits"
    data, idx_cat = _read_onehealpix_file(
        catalog_hpx['TARGETID'], fspec, arms_to_keep, skip_resomat)

    if skip_resomat:
        return data, idx_cat

    fspec = f"{input_dir}/{pixnum//100}/{pixnum}/truth-{nside}-{pixnum}.fits"
    fitsfile = fitsio.FITS(fspec)
    for arm in arms_to_keep:
        data['reso'][arm] = np.array(fitsfile[f'{arm}_RESOLUTION'].read())
    fitsfile.close()

    return data, idx_cat


def _float_range(f1, f2):
    # Define the function with default arguments
    def float_range_checker(arg):
        """New Type function for argparse - a float within predefined range.
        """
        try:
            f = float(arg)
        except ValueError:
            raise argparse.ArgumentTypeError("must be a floating point number")
        if f < f1 or f > f2:
            raise argparse.ArgumentTypeError(f"must be in range [{f1}--{f2}]")
        return f

    # Return function handle to checking function
    return float_range_checker



import pytest
import os

import fitsio
import numpy as np
from numpy.lib.recfunctions import drop_fields
import numpy.testing as npt

import qsonic.catalog


class TestCatalog(object):
    def test_read_validate_adjust_column_names(self, tmp_path):
        # Test typo in SURVEY and DUMMY column
        cat_dtype = np.dtype([
            ('TARGETID', '>i8'), ('Z', '>f8'), ('TARGET_RA', '>f8'),
            ('DEC', '>f8'), ('HPXPIXEL', '>i8'), ('SURVEX', '<U4'),
            ('DUMMY', 'f4')])
        input_catalog = np.array([
            (111, 1.1, 13.1, 13.1, 1000, b'main', 1.6),
            (222, 2.1, 23.1, 23.1, 2000, b'main', 2.6)],
            dtype=cat_dtype)

        fname = tmp_path / "test_read_validate_adjust_column_names.fits"
        with fitsio.FITS(fname, 'rw', clobber=True) as fts:
            fts.write(
                input_catalog,
                extname=list(qsonic.catalog._accepted_extnames)[0])

        catalog = qsonic.catalog._read(fname)
        os.remove(fname)

        colnames = catalog.dtype.names
        assert ('DUMMY' not in colnames)
        assert ('SURVEY' not in colnames)
        assert ('RA' not in colnames)
        assert ('DEC' in colnames)
        assert ('TARGET_RA' in colnames)
        npt.assert_equal(input_catalog['TARGETID'], catalog['TARGETID'])
        npt.assert_almost_equal(input_catalog['Z'], catalog['Z'])

        catalog1 = qsonic.catalog._validate_adjust_column_names(
            catalog.copy(), is_mock=True)
        colnames = catalog1.dtype.names
        assert ('SURVEY' not in colnames)
        assert ('RA' in colnames)
        assert ('DEC' in colnames)
        assert ('TARGET_RA' not in colnames)

        expected_msg = (
            "One of these columns must be present in the catalog: SURVEY!")
        with pytest.raises(Exception, match=expected_msg):
            qsonic.catalog._validate_adjust_column_names(
                catalog, is_mock=False)

    def test_prime_catalog(self):
        cat_dtype = np.dtype([
            ('TARGETID', '>i8'), ('Z', '>f8'), ('RA', '>f8'), ('DEC', '>f8'),
            ('HPXPIXEL', '>i8'), ('SURVEY', '<U4')])
        input_catalog = np.array([
            (999, 1.1, 93.1, 93.1, 1000, b'main'),
            (666, 3.1, 63.1, 63.1, 2000, b'main'),
            (668, 3.1, 63.1, 63.1, 2000, b'main'),
            (667, 3.1, 63.1, 63.1, 2000, b'main'),
            (777, 3.0, 73.1, 73.1, 1000, b'sv1'),
            (778, 3.0, 73.1, 73.1, 1000, b'main'),
            (779, 3.0, 73.1, 73.1, 1000, b'sv1'),
            (111, 8.1, 13.1, 13.1, 1000, b'main'),
            (444, 4.1, 43.1, 43.1, 1000, b'main')],
            dtype=cat_dtype)

        nside = 64
        keep_surveys = ['sv1', 'main']
        zmin, zmax = 2.1, 7.1
        # Sort order HPXPIXEL, SURVEY, TARGETID
        expected_catalog = np.array([
            (444, 4.1, 43.1, 43.1, 1000, b'main'),
            (778, 3.0, 73.1, 73.1, 1000, b'main'),
            (777, 3.0, 73.1, 73.1, 1000, b'sv1'),
            (779, 3.0, 73.1, 73.1, 1000, b'sv1'),
            (666, 3.1, 63.1, 63.1, 2000, b'main'),
            (667, 3.1, 63.1, 63.1, 2000, b'main'),
            (668, 3.1, 63.1, 63.1, 2000, b'main')],
            dtype=cat_dtype)
        catalog1 = qsonic.catalog._prime_catalog(
            input_catalog.copy(), nside, keep_surveys, zmin, zmax)
        npt.assert_array_equal(catalog1, expected_catalog)

        # Sort order HPXPIXEL, TARGETID
        input_catalog = drop_fields(input_catalog, 'SURVEY')
        expected_catalog = np.array([
            (444, 4.1, 43.1, 43.1, 1000),
            (777, 3.0, 73.1, 73.1, 1000),
            (778, 3.0, 73.1, 73.1, 1000),
            (779, 3.0, 73.1, 73.1, 1000),
            (666, 3.1, 63.1, 63.1, 2000),
            (667, 3.1, 63.1, 63.1, 2000),
            (668, 3.1, 63.1, 63.1, 2000)],
            dtype=input_catalog.dtype)
        catalog1 = qsonic.catalog._prime_catalog(
            input_catalog.copy(), nside, keep_surveys, zmin, zmax)
        npt.assert_array_equal(catalog1, expected_catalog)

    def test_mpi_get_local_queue(self):
        cat_dtype = np.dtype([
            ('TARGETID', '>i8'), ('Z', '>f8'), ('RA', '>f8'), ('DEC', '>f8'),
            ('HPXPIXEL', '>i8'), ('SURVEY', '<U4')])

        # Must be sorted in HPXPIXEL
        input_catalog = np.array([
            (444, 4.1, 43.1, 43.1, 1000, b'main'),
            (778, 3.0, 73.1, 73.1, 1001, b'main'),
            (777, 3.0, 73.1, 73.1, 1001, b'sv1'),
            (779, 3.0, 73.1, 73.1, 1003, b'sv1'),
            (666, 3.1, 63.1, 63.1, 2000, b'main'),
            (667, 3.1, 63.1, 63.1, 2000, b'main'),
            (668, 3.1, 63.1, 63.1, 2000, b'main')],
            dtype=cat_dtype)

        mpi_size = 2
        q0 = qsonic.catalog.mpi_get_local_queue(input_catalog, 0, mpi_size)
        q1 = qsonic.catalog.mpi_get_local_queue(input_catalog, 1, mpi_size)
        assert (len(q0) == 2)
        assert (len(q1) == 2)
        npt.assert_equal(q0[0]['TARGETID'], [666, 667, 668])
        npt.assert_equal(q1[0]['TARGETID'], [778, 777])



import argparse
import pytest
from unittest import TestCase

import numpy as np
import numpy.testing as npt

from qsonic import QsonicException
from qsonic.io import add_io_parser
import qsonic.mpi_utils
import qsonic.spectrum


class TestMPIUtils(TestCase):
    @pytest.mark.mpi(min_size=2)
    def test_mpi_parse(self):
        from mpi4py import MPI
        comm = MPI.COMM_WORLD
        assert comm.size > 0

        mpi_rank = comm.Get_rank()
        parser = argparse.ArgumentParser()
        add_io_parser(parser)

        options = "--input-dir indir --catalog incat -o outdir".split(' ')
        args = qsonic.mpi_utils.mpi_parse(parser, comm, mpi_rank, options)
        assert (args.input_dir == "indir")
        assert (args.catalog == "incat")
        assert (args.outdir == "outdir")

        with pytest.raises(SystemExit):
            options = "--catalog incat -o outdir".split(' ')
            qsonic.mpi_utils.mpi_parse(parser, comm, mpi_rank, options)

    def test_logging_mpi(self):
        with self.assertLogs(level='INFO') as cm:
            qsonic.mpi_utils.logging_mpi("test1", 0)
            qsonic.mpi_utils.logging_mpi("test2", 1)
            qsonic.mpi_utils.logging_mpi("test3", 0, "error")
        self.assertEqual(cm.output, ["INFO:root:test1", "ERROR:root:test3"])

    def test_balance_load(self):
        split_catalog = [
            np.ones(3), 2 * np.ones(4), 3 * np.ones(5), 4 * np.ones(1)]

        sorted_catalog = [
            3 * np.ones(5), 2 * np.ones(4), np.ones(3), 4 * np.ones(1)]

        mpi_size = 3
        q0 = qsonic.mpi_utils.balance_load(split_catalog, mpi_size, 0)
        q1 = qsonic.mpi_utils.balance_load(split_catalog, mpi_size, 1)
        q2 = qsonic.mpi_utils.balance_load(split_catalog, mpi_size, 2)
        for idx in range(len(split_catalog)):
            npt.assert_allclose(split_catalog[idx], sorted_catalog[idx])
        assert (len(q0) == 1)
        assert (len(q1) == 1)
        assert (len(q2) == 2)
        npt.assert_allclose(q0, 3)
        npt.assert_allclose(q1, 2)


if __name__ == '__main__':
    pytest.main()



import argparse
import os
import pytest

import fitsio
import numpy as np
import numpy.testing as npt

from qsonic.mpi_utils import mpi_parse
import qsonic.spectrum
from qsonic.picca_continuum import (
    PiccaContinuumFitter, VarLSSFitter, add_picca_continuum_parser)


@pytest.fixture
def setup_parser():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--outdir", '-o',
        help="Output directory to save deltas.")
    qsonic.spectrum.add_wave_region_parser(parser)
    parser = add_picca_continuum_parser(parser)

    return parser


@pytest.fixture
def get_fiducials(tmp_path):
    fname = tmp_path / "fiducials.fits"
    nsize = 100
    data = np.empty(
        nsize,
        dtype=[('LAMBDA', 'f8'), ('MEANFLUX', 'f8'), ('VAR', 'f8')]
    )
    data['LAMBDA'] = np.linspace(3600, 6000, nsize)
    data['MEANFLUX'] = 2 * np.ones(nsize)
    data['VAR'] = 3 * np.ones(nsize)

    with fitsio.FITS(fname, 'rw', clobber=True) as fts:
        fts.write(data, extname='STATS')

    return fname


def test_add_parser(setup_parser):
    parser = setup_parser
    parser.parse_args([])

    args = parser.parse_args(["--rfdwave", "1.2"])
    npt.assert_almost_equal(args.rfdwave, 1.2)


@pytest.mark.mpi
class TestPiccaContinuum(object):
    def test_init(self, setup_parser, get_fiducials):
        from mpi4py import MPI
        comm = MPI.COMM_WORLD
        mpi_rank = comm.Get_rank()
        mpi_size = comm.Get_size()
        assert mpi_size > 0

        parser = setup_parser
        args = mpi_parse(parser, comm, mpi_rank, [])
        qcfit = PiccaContinuumFitter(args)
        # assert rfwave is centers and forest_w1 is the edge
        w1rf_truth = args.forest_w1 + qcfit.dwrf / 2
        w2rf_truth = args.forest_w2 - qcfit.dwrf / 2

        npt.assert_allclose(
            qcfit.rfwave[[0, -1]], [w1rf_truth, w2rf_truth])
        npt.assert_almost_equal(qcfit.meancont_interp.xp0, w1rf_truth)
        npt.assert_allclose(qcfit.meancont_interp.fp, 1)
        assert (qcfit.varlss_fitter is not None)

        # Test fiducial
        if mpi_rank == 0:
            fname = str(get_fiducials)
        else:
            fname = ""

        args = mpi_parse(
            parser, comm, mpi_rank,
            ["--fiducial-meanflux", fname, "--fiducial-varlss", fname])
        qcfit = PiccaContinuumFitter(args)
        if mpi_rank == 0:
            os.remove(fname)

        npt.assert_allclose(qcfit.meanflux_interp.fp, 2)
        npt.assert_allclose(qcfit.varlss_interp.fp, 3)
        assert (qcfit.varlss_fitter is None)

    def test_get_continuum_model(self, setup_parser):
        from mpi4py import MPI
        comm = MPI.COMM_WORLD
        mpi_rank = comm.Get_rank()
        mpi_size = comm.Get_size()
        assert mpi_size > 0

        parser = setup_parser
        args = mpi_parse(parser, comm, mpi_rank, [])
        qcfit = PiccaContinuumFitter(args)

        # np.log(wave_rf_arm / self.rfwave[0]) / self._denom
        slope = np.linspace(0, 1, 200)
        wave_rf_arm = qcfit.rfwave[0] * np.exp(qcfit._denom * slope)
        expected_cont = np.linspace(0, 2, 200)

        x = np.array([1., 1.])
        cont_est = qcfit.get_continuum_model(x, wave_rf_arm)
        npt.assert_allclose(cont_est, expected_cont)

        x = np.array([1., 1., 0])
        cont_est = qcfit.get_continuum_model(x, wave_rf_arm)
        npt.assert_allclose(cont_est, expected_cont)

    def test_fit_continuum(self, setup_parser, setup_data):
        from mpi4py import MPI
        comm = MPI.COMM_WORLD
        mpi_rank = comm.Get_rank()
        mpi_size = comm.Get_size()
        assert mpi_size > 0

        parser = setup_parser
        args = mpi_parse(parser, comm, mpi_rank, [])
        qcfit = PiccaContinuumFitter(args)
        qcfit.varlss_interp.fp *= 0
        npt.assert_allclose(qcfit.meanflux_interp.fp, 1)

        cat_by_survey, npix, data = setup_data(1)
        data['ivar']['B'] *= 10
        data['ivar']['R'] *= 10
        spec = qsonic.spectrum.generate_spectra_list_from_data(
            cat_by_survey, data)[0]
        spec.set_forest_region(3600., 6000., args.forest_w1, args.forest_w2)

        qcfit.fit_continuum(spec)
        assert (spec.cont_params['valid'])
        npt.assert_almost_equal(spec.cont_params['x'], [2.1, 0])

    def test_project_normalize_meancont(self, setup_parser):
        from mpi4py import MPI
        comm = MPI.COMM_WORLD
        mpi_rank = comm.Get_rank()
        mpi_size = comm.Get_size()
        assert mpi_size > 0

        parser = setup_parser
        args = mpi_parse(parser, comm, mpi_rank, ["--cont-order", "3"])
        qcfit = PiccaContinuumFitter(args)
        x = np.log(qcfit.rfwave / qcfit.rfwave[0]) / qcfit._denom
        x = 2 * x - 1

        # add Legendre-like distortions
        new_meancont = 1.5 * np.ones_like(x)
        for ci in range(1, args.cont_order + 1):
            new_meancont += 0.5 / ci * x**ci

        new_meancont, _ = qcfit._project_normalize_meancont(new_meancont)
        npt.assert_allclose(new_meancont, 1, rtol=1e-3)
        npt.assert_almost_equal(new_meancont.mean(), 1, decimal=4)

    def test_update_mean_cont(self, setup_parser, setup_data):
        from mpi4py import MPI
        comm = MPI.COMM_WORLD
        mpi_rank = comm.Get_rank()
        mpi_size = comm.Get_size()
        assert mpi_size > 0

        parser = setup_parser
        # There are empty bins in the rest-frame, they should be extrapolated
        # correctly.
        args = mpi_parse(parser, comm, mpi_rank, [])
        qcfit = PiccaContinuumFitter(args)

        cat_by_survey, npix, data = setup_data(3)
        spectra_list = qsonic.spectrum.generate_spectra_list_from_data(
            cat_by_survey, data)

        for spec in spectra_list:
            spec.set_forest_region(
                3600., 6000., args.forest_w1, args.forest_w2)
            spec.cont_params['valid'] = True
            spec.cont_params['cont'] = {}
            for arm, wave_arm in spec.forestwave.items():
                spec.cont_params['cont'][arm] = np.ones_like(wave_arm)

        qcfit.update_mean_cont(spectra_list, False)
        npt.assert_allclose(qcfit.meancont_interp.fp, 1, rtol=1e-3)
        npt.assert_almost_equal(qcfit.meancont_interp.fp.mean(), 1)


class TestVarLSSFitter(object):
    def test_add(self, setup_data):
        nwbins = 4
        dwbins = 300.
        wbins_truth = 3600. + (0.5 + np.arange(nwbins)) * dwbins
        varlss_fitter = VarLSSFitter(
            3600, 4800, nwbins=nwbins, var1=1e-5, var2=2., nvarbins=3)
        npt.assert_almost_equal(varlss_fitter.dwobs, dwbins)
        npt.assert_allclose(varlss_fitter.waveobs, wbins_truth)

        cat_by_survey, npix, data = setup_data(1)
        varlss_fitter.add(
            data['wave']['B'], data['flux']['B'][0], data['ivar']['B'][0])

        empty_bins = np.s_[-5:]
        assert all(varlss_fitter._num_pixels[:5] == 0)
        assert all(varlss_fitter._num_pixels[empty_bins] == 0)
        expected_numqso = np.zeros((nwbins + 2) * 5, dtype=int)
        expected_numqso[[6, 11, 16]] = 1
        npt.assert_equal(varlss_fitter._num_qso, expected_numqso)

        varlss_fitter.add(
            data['wave']['R'], data['flux']['R'][0], data['ivar']['R'][0])
        expected_numqso[[11, 16]] = 2
        expected_numqso[21] = 1
        npt.assert_equal(varlss_fitter._num_qso, expected_numqso)

        expected_size = 3 * nwbins
        varlss_fitter._allreduce()
        npt.assert_equal(varlss_fitter.wvalid_bins.sum(), expected_size)
        npt.assert_equal(varlss_fitter.mean_delta.size, expected_size)

    def test_fit(self):
        varlss_fitter = VarLSSFitter(
            3600, 4800, nwbins=1, var1=1e-5, var2=2.)

        true_var_lss = np.array([0.1])
        nspec = 1000
        nwave = int((4800 - 3600) / 0.8) + 1
        wave = np.linspace(3600, 4800, nwave)
        RNST = np.random.default_rng()
        for jj in range(nspec):
            var_pipe = RNST.uniform(np.log10(1e-5), np.log10(2), size=nwave)
            var_pipe = 10**var_pipe
            std_gen = np.sqrt(var_pipe + true_var_lss)
            delta = RNST.normal(0, std_gen, nwave)

            varlss_fitter.add(wave, delta, 1 / var_pipe)

        fit_results, std_results = varlss_fitter.fit(
            true_var_lss, smooth=False)
        atol = 3 * std_results[0]
        rtol = atol / true_var_lss[0]
        npt.assert_allclose(fit_results, true_var_lss, rtol=rtol, atol=atol)


if __name__ == '__main__':
    pytest.main()



import logging
import warnings

import fitsio
import numpy as np

from qsonic import QsonicException


def mpi_parse(parser, comm, mpi_rank, options=None):
    """ Parse arguments on the master node, then broadcast.

    Arguments
    ---------
    parser: argparse.ArgumentParser
    comm: MPI.COMM_WORLD
        MPI comm object for bcast
    mpi_rank: int
        Rank of the MPI process.
    options: None or list, default: None
        Options to parse. None parses sys.argv
    """
    if mpi_rank == 0:
        try:
            args = parser.parse_args(options)
        except SystemExit:
            args = -1
    else:
        args = -1

    args = comm.bcast(args)
    if args == -1:
        exit(0)

    return args


def logging_mpi(msg, mpi_rank, fnc="info"):
    """ Logs only on `mpi_rank=0`.

    Arguments
    ---------
    msg: str
        Message to log.
    mpi_rank: int
        Rank of the MPI process.
    fnc: logging attr
        Channel to log, usually info or error.
    """
    if mpi_rank == 0:
        getattr(logging, fnc)(msg)


def warn_mpi(msg, mpi_rank):
    """ Warns of RuntimeWarning only on `mpi_rank=0`.

    Arguments
    ---------
    msg: str
        Message to log.
    mpi_rank: int
        Rank of the MPI process.
    """

    if mpi_rank == 0:
        warnings.warn(msg, RuntimeWarning)


def balance_load(split_catalog, mpi_size, mpi_rank):
    """ Load balancing function.

    Arguments
    ---------
    split_catalog: list(:external+numpy:py:class:`ndarray <numpy.ndarray>`)
        List of catalog. Each element is a ndarray with the same healpix.
    mpi_size: int
        Number of MPI tasks running.
    mpi_rank: int
        Rank of the MPI process.

    Returns
    ---------
    local_queue: list(:external+numpy:py:class:`ndarray <numpy.ndarray>`)
        Spectra that current rank is reponsible for in `split_catalog` format.
    """
    number_of_spectra = np.zeros(mpi_size, dtype=int)
    local_queue = []
    split_catalog.sort(key=lambda x: x.size, reverse=True)  # Descending order
    for cat in split_catalog:
        min_idx = np.argmin(number_of_spectra)
        number_of_spectra[min_idx] += cat.size

        if min_idx == mpi_rank:
            local_queue.append(cat)

    return local_queue


class MPISaver():
    """ A simple class to write to a FITS file on master node.

    Parameters
    ----------
    fname: str
        Filename. Does not create if empty string.
    mpi_rank: int
        Rank of the MPI process. Creates FITS if 0.
    """

    def __init__(self, fname, mpi_rank):
        if mpi_rank == 0 and fname:
            self.fts = fitsio.FITS(fname, 'rw', clobber=True)
        else:
            self.fts = None

    def close(self):
        """ Close FITS file."""
        if self.fts is not None:
            self.fts.close()

    def write(self, data, names=None, extname=None, header=None):
        """ Write to FITS file.

        Arguments
        ---------
        data: list(:external+numpy:py:class:`ndarray <numpy.ndarray>`)
            Data to write to extention.
        names: list(str)or None, default: None
            Column names for data.
        extname: str or None, default: None
            Extention name.
        header: dict or None, default: None
            Header dictionary to save.
        """
        if self.fts is not None:
            self.fts.write(data, names=names, extname=extname, header=header)



import os
import pytest

import numpy.testing as npt

import qsonic.masks
from qsonic.spectrum import Spectrum


def test_skymask(tmp_path, setup_data):
    # Create skymask
    fname_skymask = tmp_path / "test_skymask.txt"
    with open(fname_skymask, 'w') as file_sky:
        file_sky.write("Ca\t 3700\t 3750\t OBS\n")
        file_sky.write("RF\t 1142\t 1150\t RF\n")
        file_sky.write("Ca\t 4150\t 4200\t OBS\n")

    skymask = qsonic.masks.SkyMask(fname_skymask)
    os.remove(fname_skymask)

    # Create spectrum
    cat_by_survey, _, data = setup_data(1)
    spec = Spectrum(
        cat_by_survey[0], data['wave'], data['flux'],
        data['ivar'], data['mask'], data['reso'], 0
    )
    spec.set_forest_region(3600., 6000., 1000., 2000.)
    z_qso = cat_by_survey['Z']

    skymask.apply(spec)

    for arm, wave_arm in spec.forestwave.items():
        w = (wave_arm >= 3700.) & (wave_arm < 3750.)
        w |= (wave_arm >= 4150.) & (wave_arm < 4200.)
        w |= (wave_arm >= 1142 * (1 + z_qso)) & (wave_arm < 1150 * (1 + z_qso))
        npt.assert_equal(spec.forestivar[arm][w], 0)
        npt.assert_equal(spec.forestivar[arm][~w], 1)


if __name__ == '__main__':
    pytest.main()
